use std::{sync::Arc, time::Duration};

use managesieve::Capabilities;
use thiserror::Error;
use tokio::{
    io::{split, AsyncBufReadExt, BufReader, WriteHalf},
    net::TcpStream,
    sync::Mutex,
};
use tokio_rustls::{
    client::TlsStream,
    rustls::{ClientConfig, KeyLogFile, RootCertStore},
    TlsConnector,
};
use tracing::{error, info, trace};

pub mod parser;

#[derive(Debug, Clone)]
pub struct ConnectionInfo {
    domain: String,
    port: u16,
}

impl ConnectionInfo {
    pub fn new(domain: &str, port: u16) -> Self {
        Self {
            domain: domain.to_string(),
            port,
        }
    }
    pub async fn connect(self) -> Result<ConnectionConnected, anyhow::Error> {
        info!("connecting to the server");
        trace!("Building Cert Store");
        let mut root_store = RootCertStore::empty();
        root_store.extend(webpki_roots::TLS_SERVER_ROOTS.iter().cloned());
        let mut config = ClientConfig::builder()
            .with_root_certificates(root_store)
            .with_no_client_auth();

        // Allow using SSLKEYLOGFILE.
        config.key_log = Arc::new(KeyLogFile::new());
        trace!("Creating connector");
        let connector = TlsConnector::from(Arc::new(config));
        trace!("Connecting to the server.");
        let stream = TcpStream::connect((self.domain.clone(), self.port)).await?;
        let domain = pki_types::ServerName::try_from(self.domain.clone())?.to_owned();
        trace!("Attach the connector");
        let stream = connector.connect(domain, stream).await?;
        let (reader, writer) = split(stream);
        let buffer = Arc::new(Mutex::new(String::new()));
        let spawned_buffer = buffer.clone();

        let _join_handle = tokio::spawn(async move {
            let mut buf = String::new();
            let mut reader = BufReader::new(reader);
            loop {
                trace!("Waiting for more data to read…");
                let read = reader.read_line(&mut buf).await;
                if let Ok(num) = read {
                    trace!("Read {} bytes", num);
                    trace!("Read Buffer: {:?}", &buf);
                    {
                        let mut s = spawned_buffer.lock().await;
                        s.push_str(&buf);
                    }
                    buf.clear();
                };
                tokio::time::sleep(Duration::from_millis(50)).await;
            }
        });

        let caps = loop {
            let mut buf = buffer.lock().await;
            trace!("reading new input:\n{}", buf);
            match managesieve::response_capability(&buf.to_string()) {
                Ok((rest, caps, resp)) => match resp.tag {
                    managesieve::OkNoBye::Ok => {
                        buf.clear();
                        buf.push_str(rest);
                        info!("Read the introduction");
                        break caps;
                    }
                    managesieve::OkNoBye::No | managesieve::OkNoBye::Bye => {
                        trace!("Connection closed!");
                        //panic!("Invalid Response")
                    }
                },
                Err(e) => trace!("(temporary) error: {}", e),
            }
            tokio::time::sleep(Duration::from_millis(50)).await;
        };
        trace!("Capabilities read: {:#?}", &caps);
        Ok(ConnectionConnected {
            info: self,
            writer: Arc::new(writer),
            buffer,
            server_settings: caps,
        })
    }
}

#[derive(Debug, Clone)]
pub struct ConnectionConnected {
    info: ConnectionInfo,
    //reader: Arc<ReadHalf<TlsStream<TcpStream>>>,
    writer: Arc<WriteHalf<TlsStream<TcpStream>>>,
    buffer: Arc<Mutex<String>>,
    //pub join_handle: Arc<JoinHandle<()>>,
    server_settings: Capabilities,
}

#[derive(Error, Debug)]
pub enum ReadError {
    #[error("The data entered was invalid")]
    InvalidInput(#[source] managesieve::Error),
    #[error("The data the server responded was invalid")]
    InvalidResponse(#[source] managesieve::Error),
}
#[derive(Debug)]
pub enum IsComplete<T> {
    Yes(T),
    No(T),
}

impl ConnectionConnected {
    pub async fn log_buffer(&self) {
        info!("Buffer is: \n{}", self.buffer.lock().await)
    }
    pub async fn log_server_settings(&self) {
        info!("Serversettings:\n{:?}", self.server_settings)
    }
    #[tracing::instrument(name = "gettin the intro")]
    pub async fn read_introduction(self) -> Result<IsComplete<Self>, anyhow::Error> {
        let Self {
            info,
            //reader: Arc<ReadHalf<TlsStream<TcpStream>>>,
            writer,
            buffer,
            server_settings,
        } = self;
        tokio::time::sleep(Duration::from_millis(50)).await;
        let mut bf = buffer.lock().await;
        let response = match managesieve::response_capability(&bf.to_string()) {
            Ok((rest, capability, response)) => {
                info!("Successfully read the introduction {:?}", response);
                bf.clear();
                bf.push_str(rest);
                drop(bf);
                Ok(IsComplete::Yes(Self {
                    info,
                    writer: writer.clone(),
                    buffer: buffer.clone(),
                    server_settings: capability,
                }))
            }
            Err(managesieve::Error::IncompleteResponse) => {
                trace!("incomplete introduction");
                Ok(IsComplete::No(Self {
                    info,
                    writer,
                    buffer: buffer.clone(),
                    server_settings,
                }))
            }
            Err(managesieve::Error::InvalidResponse) => {
                error!("invalid response");
                Err(managesieve::Error::InvalidResponse)?
            }
            Err(managesieve::Error::InvalidInput) => {
                error!("invalid input");
                Err(managesieve::Error::InvalidResponse)?
            }
            Err(managesieve::Error::MissingLine(line)) => {
                error!("a capability line was missing: {:?}", line);
                Err(managesieve::Error::InvalidResponse)?
            }
        };

        response
    }

    pub fn get_greeting(&self) -> String {
        self.server_settings.implementation.to_string()
    }
    pub fn get_server_capabilities(&self) -> &Capabilities {
        &self.server_settings
    }
}

#[derive(Debug, Clone)]
pub struct ConnectionAuthenticated {
    info: ConnectionInfo,
    server_settings: ServerSettings,
}

#[derive(Debug, Clone)]
pub struct ServerSettings {
    pub implementation: String,
    pub version: Vec<i32>,
    pub sasl: Vec<Methods>,
    pub capabilities: Vec<String>,
    pub notify: Vec<String>,
    pub max_redirects: u32,
}

#[derive(Debug, Clone, Copy)]
pub enum Methods {
    Plain,
    OAuthbearer,
}
